@@ -1,4 +1,5 @@
 import os
+import sys
 import hashlib
 import time
 import queue
@@ -8,14 +9,31 @@ import getpass
 import logging
 import base64
 import threading
-from typing import Any, Dict, List, Type, Callable, Optional, DefaultDict, Tuple, Union
+from typing import (
+    Any,
+    Dict,
+    List,
+    Type,
+    Callable,
+    Optional,
+    DefaultDict,
+    Union,
+    Tuple,
+)
 from types import FrameType
 from collections import defaultdict
+import enum
 
 from telegram import VERSION
 from telegram.utils import AsyncResult
 from telegram.tdjson import TDJson
 from telegram.worker import BaseWorker, SimpleWorker
+from telegram.text import Element
+
+if sys.version_info >= (3, 8):  # Backwards compatibility for python < 3.8
+    from typing import Literal
+else:
+    from typing_extensions import Literal
 
 logger = logging.getLogger(__name__)
 
@@ -23,6 +41,19 @@ logger = logging.getLogger(__name__)
 MESSAGE_HANDLER_TYPE: str = 'updateNewMessage'
 
 
+class AuthorizationState(enum.Enum):
+    NONE = None
+    WAIT_CODE = 'authorizationStateWaitCode'
+    WAIT_PASSWORD = 'authorizationStateWaitPassword'
+    WAIT_TDLIB_PARAMETERS = 'authorizationStateWaitTdlibParameters'
+    WAIT_ENCRYPTION_KEY = 'authorizationStateWaitEncryptionKey'
+    WAIT_PHONE_NUMBER = 'authorizationStateWaitPhoneNumber'
+    WAIT_REGISTRATION = 'authorizationStateWaitRegistration'
+    READY = 'authorizationStateReady'
+    CLOSING = 'authorizationStateClosing'
+    CLOSED = 'authorizationStateClosed'
+
+
 class Telegram:
     def __init__(
         self,
@@ -80,6 +111,7 @@ class Telegram:
         self.proxy_port = proxy_port
         self.proxy_type = proxy_type
         self.use_secret_chats = use_secret_chats
+        self.authorization_state = AuthorizationState.NONE
 
         if not self.bot_token and not self.phone:
             raise ValueError('You must provide bot_token or phone')
@@ -96,16 +128,14 @@ class Telegram:
         self.files_directory = files_directory
 
         self._authorized = False
-        self._is_enabled = False
+        self._stopped = threading.Event()
 
         # todo: move to worker
-        self._workers_queue: queue.Queue = queue.Queue(
-            maxsize=default_workers_queue_size
-        )
+        self._workers_queue: queue.Queue = queue.Queue(maxsize=default_workers_queue_size)
 
         if not worker:
             worker = SimpleWorker
-        self.worker = worker(queue=self._workers_queue)
+        self.worker: BaseWorker = worker(queue=self._workers_queue)
 
         self._results: Dict[str, AsyncResult] = {}
         self._update_handlers: DefaultDict[str, List[Callable]] = defaultdict(list)
@@ -116,17 +146,86 @@ class Telegram:
         if login:
             self.login()
 
-    def __del__(self) -> None:
-        self.stop()
-
     def stop(self) -> None:
         """Stops the client"""
-        self._is_enabled = False
+
+        if self._stopped.is_set():
+            return
+
+        logger.info('Stopping telegram client...')
+
+        self._close()
+        self.worker.stop()
+        self._stopped.set()
+
+        # wait for the tdjson listener to stop
+        self._td_listener.join()
 
         if hasattr(self, '_tdjson'):
             self._tdjson.stop()
 
-    def send_message(self, chat_id: int, text: str) -> AsyncResult:
+    def _close(self) -> None:
+        """
+        Calls `close` tdlib method and waits until authorization_state becomes CLOSED.
+        Blocking.
+        """
+        self.call_method('close')
+
+        while self.authorization_state != AuthorizationState.CLOSED:
+            result = self.get_authorization_state()
+            self.authorization_state = self._wait_authorization_result(result)
+            logger.info('Authorization state: %s', self.authorization_state)
+            time.sleep(0.5)
+
+    def parse_text_entities(self, text: str, parse_mode: Literal['HTML', 'Markdown']) -> AsyncResult:
+        """
+        Parses text from 'HTML' and 'Markdown' (not MarkdownV2) into plain
+        text and internal telegram style description.
+
+        Args:
+            text
+            parse_mode
+
+        Returns:
+            AsyncResult
+            The update will be:
+                {
+                    '@type': 'formattedText',
+                    'text': 'Hello world!',
+                    'entities': [
+                        {
+                            '@type': 'textEntity',
+                            'offset': 0,
+                            'length': 12,
+                            'type': {
+                                '@type': 'textEntityTypeSpoiler'
+                            }
+                        }
+                        ...
+                    ]
+                }
+        """
+
+        parse_mode_types = {
+            'HTML': 'textParseModeHTML',
+            'Markdown': 'textParseModeMarkdown',
+        }
+        data = {
+            '@type': 'parseTextEntities',
+            'text': text,
+            'parse_mode': {
+                '@type': parse_mode_types[parse_mode],
+            },
+        }
+
+        return self._send_data(data)
+
+    def send_message(
+        self,
+        chat_id: int,
+        text: Union[str, Element],
+        entities: Union[List[dict], None] = None,
+    ) -> AsyncResult:
         """
         Sends a message to a chat. The chat must be in the tdlib's database.
         If there is no chat in the DB, tdlib returns an error.
@@ -147,17 +246,81 @@ class Telegram:
                     ...
                 }
         """
+
+        if entities is None:
+            entities = []
+
+        updated_text: str
+        if isinstance(text, Element):
+            result = self.parse_text_entities(text.to_html(), parse_mode='HTML')
+            result.wait()
+            assert result.update is not None
+            update: dict = result.update
+            entities = update['entities']
+            updated_text = update['text']
+        else:
+            updated_text = text
+
         data = {
             '@type': 'sendMessage',
             'chat_id': chat_id,
             'input_message_content': {
                 '@type': 'inputMessageText',
-                'text': {'@type': 'formattedText', 'text': text},
+                'text': {
+                    '@type': 'formattedText',
+                    'text': updated_text,
+                    'entities': entities,
+                },
             },
         }
 
         return self._send_data(data)
 
+    def import_contacts(self, contacts: List[Dict[str, str]]) -> AsyncResult:
+        """
+        Adds new contacts or edits existing contacts by their phone numbers.
+        https://core.telegram.org/tdlib/docs/classtd_1_1td__api_1_1import_contacts.html
+
+        Args:
+            contacts
+
+        contacts is a list of the form
+            [
+                {
+                    "phone_number": "+380 12 345 67 89",
+                    "first_name": "Name",
+                    "last_name": "Surname"
+                },
+                {
+                    "phone_number": "+380 09 876 54 32",
+                    "first_name": "Name",
+                    "last_name": "Surname"
+                },
+                ...
+            ]
+        phone format is country-specifc
+
+        Returns:
+            AsyncResult
+            The update will be:
+            {
+                '@type': 'importedContacts',
+                'user_ids': [1, 2],
+                'importer_count': [3, 4],
+                ...
+            }
+        """
+
+        for contact in contacts:
+            contact["@type"] = "contact"
+
+        data = {
+            '@type': 'importContacts',
+            'contacts': contacts,
+        }
+
+        return self._send_data(data)
+
     def get_chat(self, chat_id: int) -> AsyncResult:
         """
         This is offline request, if there is no chat in your database it will not be found
@@ -185,9 +348,16 @@ class Telegram:
 
         return self.call_method('getUser', params={'user_id': user_id})
 
-    def get_chats(
-        self, offset_order: int = 0, offset_chat_id: int = 0, limit: int = 100
-    ) -> AsyncResult:
+    def get_user_full_info(self, user_id: int) -> AsyncResult:
+        """
+        Requests the full information about a user with id = user_id.
+
+        https://core.telegram.org/tdlib/docs/classtd_1_1td__api_1_1get_user_full_info.html
+        """
+
+        return self.call_method('getUserFullInfo', params={'user_id': user_id})
+
+    def get_chats(self, offset_order: int = 0, offset_chat_id: int = 0, limit: int = 100) -> AsyncResult:
         """
         Returns a list of chats:
 
@@ -238,7 +408,11 @@ class Telegram:
 
         return self._send_data(data)
 
-    def get_message(self, chat_id: int, message_id: int,) -> AsyncResult:
+    def get_message(
+        self,
+        chat_id: int,
+        message_id: int,
+    ) -> AsyncResult:
         """
         Return a message via its message_id
 
@@ -263,11 +437,10 @@ class Telegram:
             'chat_id': chat_id,
             'message_id': message_id,
         }
+
         return self._send_data(data)
 
-    def delete_messages(
-        self, chat_id: int, message_ids: List[int], revoke: bool = True
-    ) -> AsyncResult:
+    def delete_messages(self, chat_id: int, message_ids: List[int], revoke: bool = True) -> AsyncResult:
         """
         Delete a list of messages in a chat
 
@@ -276,6 +449,7 @@ class Telegram:
             message_ids
             revoke
         """
+
         return self._send_data(
             {
                 '@type': 'deleteMessages',
@@ -292,9 +466,8 @@ class Telegram:
         Args:
             supergroup_id
         """
-        return self._send_data(
-            {'@type': 'getSupergroupFullInfo', 'supergroup_id': supergroup_id}
-        )
+
+        return self._send_data({'@type': 'getSupergroupFullInfo', 'supergroup_id': supergroup_id})
 
     def create_basic_group_chat(self, basic_group_id: int) -> AsyncResult:
         """
@@ -303,13 +476,10 @@ class Telegram:
         Args:
             basic_group_id
         """
-        return self._send_data(
-            {'@type': 'createBasicGroupChat', 'basic_group_id': basic_group_id}
-        )
 
-    def get_web_page_instant_view(
-        self, url: str, force_full: bool = False
-    ) -> AsyncResult:
+        return self._send_data({'@type': 'createBasicGroupChat', 'basic_group_id': basic_group_id})
+
+    def get_web_page_instant_view(self, url: str, force_full: bool = False) -> AsyncResult:
         """
         Use this method to request instant preview of a webpage.
         Returns error with 404 if there is no preview for this webpage.
@@ -343,8 +513,6 @@ class Telegram:
         return self._send_data(data, block=block)
 
     def _run(self) -> None:
-        self._is_enabled = True
-
         self._td_listener = threading.Thread(target=self._listen_to_td)
         self._td_listener.daemon = True
         self._td_listener.start()
@@ -354,21 +522,17 @@ class Telegram:
     def _listen_to_td(self) -> None:
         logger.info('[Telegram.td_listener] started')
 
-        while self._is_enabled:
+        while not self._stopped.is_set():
             update = self._tdjson.receive()
 
             if update:
                 self._update_async_result(update)
                 self._run_handlers(update)
 
-    def _update_async_result(
-        self, update: Dict[Any, Any]
-    ) -> typing.Optional[AsyncResult]:
+    def _update_async_result(self, update: Dict[Any, Any]) -> typing.Optional[AsyncResult]:
         async_result = None
 
-        _special_types = (
-            'updateAuthorizationState',
-        )  # for authorizationProcess @extra.request_id doesn't work
+        _special_types = ('updateAuthorizationState',)  # for authorizationProcess @extra.request_id doesn't work
 
         if update.get('@type') in _special_types:
             request_id = update['@type']
@@ -381,11 +545,10 @@ class Telegram:
             async_result = self._results.get(request_id)
 
         if not async_result:
-            logger.debug(
-                'async_result has not been found in by request_id=%s', request_id
-            )
+            logger.debug('async_result has not been found in by request_id=%s', request_id)
         else:
             done = async_result.parse_update(update)
+
             if done:
                 self._results.pop(request_id, None)
 
@@ -393,6 +556,7 @@ class Telegram:
 
     def _run_handlers(self, update: Dict[Any, Any]) -> None:
         update_type: str = update.get('@type', 'unknown')
+
         for handler in self._update_handlers[update_type]:
             self._workers_queue.put((handler, update), timeout=self._queue_put_timeout)
 
@@ -424,6 +588,7 @@ class Telegram:
 
         If `block`is True, waits for the result
         """
+
         if '@extra' not in data:
             data['@extra'] = {}
 
@@ -442,20 +607,27 @@ class Telegram:
         return async_result
 
     def idle(
-        self, stop_signals: Tuple = (signal.SIGINT, signal.SIGTERM, signal.SIGABRT)
+        self,
+        stop_signals: Tuple = (
+            signal.SIGINT,
+            signal.SIGTERM,
+            signal.SIGABRT,
+            signal.SIGQUIT,
+        ),
     ) -> None:
-        """Blocks until one of the signals are received and stops"""
+        """
+        Blocks until one of the exit signals is received.
+        When a signal is received, calls `stop`.
+        """
 
         for sig in stop_signals:
-            signal.signal(sig, self._signal_handler)
+            signal.signal(sig, self._stop_signal_handler)
 
-        self._is_enabled = True
+        self._stopped.wait()
 
-        while self._is_enabled:
-            time.sleep(0.1)
-
-    def _signal_handler(self, signum: int, frame: FrameType) -> None:
-        self._is_enabled = False
+    def _stop_signal_handler(self, signum: int, frame: Optional[FrameType] = None) -> None:
+        logger.info('Signal %s received!', signum)
+        self.stop()
 
     def get_authorization_state(self) -> AsyncResult:
         logger.debug('Getting authorization state')
@@ -463,45 +635,87 @@ class Telegram:
 
         return self._send_data(data, result_id='getAuthorizationState')
 
-    def login(self) -> None:
+    def _wait_authorization_result(self, result: AsyncResult) -> AuthorizationState:
+        authorization_state = None
+
+        if result:
+            result.wait(raise_exc=True)
+
+            if result.update is None:
+                raise RuntimeError('Something wrong, the result update is None')
+
+            if result.id == 'getAuthorizationState':
+                authorization_state = result.update['@type']
+            else:
+                authorization_state = result.update['authorization_state']['@type']
+
+        return AuthorizationState(authorization_state)
+
+    def login(self, blocking: bool = True) -> AuthorizationState:
         """
-        Login process (blocking)
+        Login process.
 
         Must be called before any other call.
         It sends initial params to the tdlib, sets database encryption key, etc.
+
+        args:
+          blocking [bool]: If True, the process is blocking and the client
+                           expects password and code from stdin.
+                           If False, `login` call returns next AuthorizationState and
+                           the login process can be continued (with calling login(blocking=False) again)
+                           after the necessary action is completed.
+
+        Returns:
+         - AuthorizationState.WAIT_CODE if a telegram code is required.
+           The caller should ask the telegram code
+           to the end user then call send_code(code)
+         - AuthorizationState.WAIT_PASSWORD if a telegram password is required.
+           The caller should ask the telegram password
+           to the end user and then call send_password(password)
+         - AuthorizationState.WAIT_REGISTRATION if a the user must finish registration
+           The caller should ask the first and last names
+           to the end user and then call register_user(first, last)
+         - AuthorizationState.READY if the login process succeeded.
         """
+
         if self.proxy_server:
             self._send_add_proxy()
 
-        authorization_state = None
-        actions = {
-            None: self.get_authorization_state,
-            'authorizationStateWaitTdlibParameters': self._set_initial_params,
-            'authorizationStateWaitEncryptionKey': self._send_encryption_key,
-            'authorizationStateWaitPhoneNumber': self._send_phone_number_or_bot_token,
-            'authorizationStateWaitCode': self._send_telegram_code,
-            'authorizationStateWaitPassword': self._send_password,
-            'authorizationStateReady': self._complete_authorization,
+        actions: Dict[AuthorizationState, Callable[[], AsyncResult]] = {
+            AuthorizationState.NONE: self.get_authorization_state,
+            AuthorizationState.WAIT_TDLIB_PARAMETERS: self._set_initial_params,
+            AuthorizationState.WAIT_ENCRYPTION_KEY: self._send_encryption_key,
+            AuthorizationState.WAIT_PHONE_NUMBER: self._send_phone_number_or_bot_token,
+            AuthorizationState.WAIT_CODE: self._send_telegram_code,
+            AuthorizationState.WAIT_PASSWORD: self._send_password,
+            AuthorizationState.WAIT_REGISTRATION: self._register_user,
         }
+
+        blocking_actions = (
+            AuthorizationState.WAIT_CODE,
+            AuthorizationState.WAIT_PASSWORD,
+            AuthorizationState.WAIT_REGISTRATION,
+        )
+
         if self.phone:
             logger.info('[login] Login process has been started with phone')
         else:
             logger.info('[login] Login process has been started with bot token')
 
-        while not self._authorized:
-            logger.info('[login] current authorization state: %s', authorization_state)
-            result = actions[authorization_state]()
+        while self.authorization_state != AuthorizationState.READY:
+            logger.info('[login] current authorization state: %s', self.authorization_state)
 
-            if result:
-                result.wait(raise_exc=True)
+            if not blocking and self.authorization_state in blocking_actions:
+                return self.authorization_state
 
-                if result.update is None:
-                    raise RuntimeError('Something wrong, the result update is None')
+            result = actions[self.authorization_state]()
 
-                if result.id == 'getAuthorizationState':
-                    authorization_state = result.update['@type']
-                else:
-                    authorization_state = result.update['authorization_state']['@type']
+            if not isinstance(result, AuthorizationState):
+                self.authorization_state = self._wait_authorization_result(result)
+            else:
+                self.authorization_state = result
+
+        return self.authorization_state
 
     def _set_initial_params(self) -> AsyncResult:
         logger.info(
@@ -533,6 +747,7 @@ class Telegram:
         logger.info('Sending encryption key')
 
         key = self._database_encryption_key
+
         if isinstance(key, str):
             key = key.encode()
 
@@ -545,6 +760,7 @@ class Telegram:
 
     def _send_phone_number_or_bot_token(self) -> AsyncResult:
         """Sends phone number or a bot_token"""
+
         if self.phone:
             return self._send_phone_number()
         elif self.bot_token:
@@ -572,6 +788,7 @@ class Telegram:
             'enable': True,
             'type': self.proxy_type,
         }
+
         return self._send_data(data, result_id='setProxy')
 
     def _send_bot_token(self) -> AsyncResult:
@@ -580,20 +797,96 @@ class Telegram:
 
         return self._send_data(data, result_id='updateAuthorizationState')
 
-    def _send_telegram_code(self) -> AsyncResult:
+    def _send_telegram_code(self, code: Optional[str] = None) -> AsyncResult:
         logger.info('Sending code')
-        code = input('Enter code:')
+
+        if code is None:
+            code = input('Enter code:')
         data = {'@type': 'checkAuthenticationCode', 'code': str(code)}
 
         return self._send_data(data, result_id='updateAuthorizationState')
 
-    def _send_password(self) -> AsyncResult:
+    def send_code(self, code: str) -> AuthorizationState:
+        """
+        Verifies a telegram code and continues the authorization process
+
+        Args:
+          code: the code to be verified. If code is None, it will be asked to the user using the input() function
+
+        Returns
+         - AuthorizationState. The called have to call `login` to continue the login process.
+
+        Raises:
+         - RuntimeError if the login failed
+        """
+        result = self._send_telegram_code(code)
+        self.authorization_state = self._wait_authorization_result(result)
+
+        return self.authorization_state
+
+    def _send_password(self, password: Optional[str] = None) -> AsyncResult:
         logger.info('Sending password')
-        password = getpass.getpass('Password:')
+
+        if password is None:
+            password = getpass.getpass('Password:')
         data = {'@type': 'checkAuthenticationPassword', 'password': password}
 
         return self._send_data(data, result_id='updateAuthorizationState')
 
-    def _complete_authorization(self) -> None:
-        logger.info('Completing auth process')
-        self._authorized = True
+    def send_password(self, password: str) -> AuthorizationState:
+        """
+        Verifies a telegram password and continues the authorization process
+
+        Args:
+          password the password to be verified.
+          If password is None, it will be asked to the user using the getpass.getpass() function
+
+        Returns
+         - AuthorizationState. The called have to call `login` to continue the login process.
+
+        Raises:
+          - RuntimeError if the login failed
+
+        """
+        result = self._send_password(password)
+        self.authorization_state = self._wait_authorization_result(result)
+
+        return self.authorization_state
+
+    def _register_user(self, first: Optional[str] = None, last: Optional[str] = None) -> AsyncResult:
+        logger.info('Registering user')
+
+        if first is None:
+            first = input('Enter first name: ')
+
+        if last is None:
+            last = input('Enter last name: ')
+
+        data = {
+            '@type': 'registerUser',
+            'first_name': first,
+            'last_name': last,
+        }
+
+        return self._send_data(data, result_id='updateAuthorizationState')
+
+    def register_user(self, first: str, last: str) -> AuthorizationState:
+        """
+        Finishes the new user registration process
+
+        Args:
+          first the user's first name
+          last the user's last name
+          If either argument is None, it will be asked to the user using the input() function
+
+        Returns
+         - AuthorizationState. The called have to call `login` to continue the login process.
+
+        Raises:
+          - RuntimeError if the login failed
+
+        """
+        result = self._register_user(first, last)
+        self.authorization_state = self._wait_authorization_result(result)
+
+        return self.authorization_state
