@@ -1,6 +1,7 @@
 import json
 import logging
 import platform
+import ctypes.util
 from ctypes import CDLL, CFUNCTYPE, c_int, c_char_p, c_double, c_void_p, c_longlong
 from typing import Any, Dict, Optional, Union
 
@@ -10,6 +11,11 @@ logger = logging.getLogger(__name__)
 
 
 def _get_tdjson_lib_path() -> str:
+    system_library = ctypes.util.find_library("tdjson")
+
+    if system_library is not None:
+        return system_library
+
     if platform.system().lower() == 'darwin':
         lib_name = 'darwin/libtdjson.dylib'
     else:
@@ -94,6 +100,7 @@ class TDJson:
         if result_str:
             result: Dict[Any, Any] = json.loads(result_str.decode('utf-8'))
             logger.debug('[me <==] Received %s', result)
+
             return result
 
         return None
@@ -104,6 +111,7 @@ class TDJson:
 
         if result_str:
             result: Dict[Any, Any] = json.loads(result_str.decode('utf-8'))
+
             return result
 
         return None
